#!/usr/local/bin/ruby
# Copyright (c) 2001 Brian Fundakowski Feldman
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: bonkle.rb,v 1.14 2001/03/01 01:08:26 green Exp $
require 'devrandom'

# This is Bonkle, a game where 16 dice with letters are randomly placed
# in a 4x4 board, then words are matched using consecutive dice.
module Bonkle
	# Pull random dice (on random sides) from a defined set of dice
	class RandomDice
		# Get a set of dice.
		# This will raise an exception if DevRandom cannot be initialized.
		def initialize(width = 4)
			@dr = DevRandom.new
			@width = width
			reset
			self
		end
		# Reinitialize to a full set of dice
		def reset
			# Allow for different board sizes.
			case @width
			when 4	# "standard"
				@dice = [
					%w(A A E E G N),
					%w(A B B J O O),
					%w(A C H O P S),
					%w(A F F K P S),
					%w(A O O T T W),
					%w(C I M O T U),
					%w(D E I L R X),
					%w(D E L R V Y),
					%w(D I S T T Y),
					%w(E I O S S T),
					%w(E E I N S U),
					%w(E E G H N W),
					%w(E H R T V W),
					%w(E L R T T Y),
					%w(H I M N Qu U),
					%w(H L N N R Z),
				]
			when 5	# "big"
				raise ArgumentError, '"big" Bonkle not yet implemented fully'
			else
				raise ArgumentError, 'invalid size (should be 4-5)'
			end
			true
		end	
		# Take random dice from the set.  With no arguments or passed 1,
		# return the single string from one die, else an Array of strings
		# of the number of dice asked for.
		#
		# This raises an EOFError exception if too many dice are asked for.
		def grab(count = 1)
			if @dice.size == 0
				raise EOFError, 'selection of dice empty'
			end
			case count
			when 2 .. @dice.size
				ret = []
				count.downto(1) {
					pos = @dr.read(2) % @dice.size
					die = @dice.delete_at(pos)
					ret.push die[@dr.read(2) % 6]
				}
				return ret
			when 1
				pos = @dr.read(2) % @dice.size
				die = @dice.delete_at(pos)
				return die[@dr.read(2) % 6]
			else
				raise ArgumentError, 'bad die count'
			end
		end
		# Are there dice left?
		def empty?
			@dice.empty?
		end
		# Are there all the dice left?
		def full?
			@dice.size == @width * @width
		end
	end
	# This represents the actual gaming space and the matching of words
	class Board
		# Create a new "board" of random dice
		# initialize([width,] [initial_board])
		def initialize(*args)
			if args[0].is_a?(Numeric)
				@width = args[0]
				@initial_board = args[1]
			else
				@initial_board = args[0]
				@width = @initial_board ? @initial_board.size : 4
			end
			if !@initial_board
				@dice = Bonkle::RandomDice.new(@width)
			end
			reset
			self
		end
		attr_reader :width
		# Reinitialize the board to a new random one or initially given one
		def reset
			if !@initial_board
				@board = Internal.new
				@dice.reset if !@dice.full?
				@width.times {
					@board.push @dice.grab(@width)
				}
			else
				if @initial_board.is_a?(Bonkle::Board::Internal)
					@board = @initial_board.dup
				else
					if !@initial_board.is_a?(Enumerable) ||
					   @initial_board.size != @width ||
					   @initial_board.find {|row|
						   !row.is_a?(Enumerable) ||
						   row.size != @width ||
					       row.find {|letter| letter !~ /^[A-PR-Z]|Qu$/i}
					   }
						raise ArgumentError, "invalid initial board"
					end
					@board = Internal.new
					@width.times {|rowi|
						@board.push @initial_board[rowi]
					}
					# detach from the original's rows and entries
					@board = @board.dup
					# to be friendly, change all letters to [A-PR-Z]|Qu
					@board.each {|row|
						row.each {|die|
							die.upcase!
							die.gsub!(/QU/, 'Qu')
						}
					}
				end
			end
			freeze_board
			true
		end
		# Fortify a standard Array with methods to make it a Bonkle "matrix"
		class Internal < Array
			def find_letter(letter)
				if $DEBUG
					printf("find_letter(\"%s\")\n", letter)
				end
				locations = []
				self.each_with_index {|row, rowi|
					row.each_with_index {|die, diei|
						if letter == die
							locations.push [rowi, diei]
						end
					}
				}
				locations
			end
			# Take an array of letters and find the next one.
			def find_word(word, prev_coords)
				if $DEBUG
					printf("find_word(\"%s\", [%d, %d])\n", word, *prev_coords)
				end
				return true if word.empty?
				self.find_letter(word[0]).each {|coords|
					if (coords[0] - prev_coords[0]).between?(-1, 1) &&
					   (coords[1] - prev_coords[1]).between?(-1, 1)
						self[*coords].concat '_'
						return true if self.find_word(word[1..-1], coords)
						self[*coords].chop!
					end
				}
				false
			end
			def dup
				b = super
				b.each_index {|rowi|
					b[rowi] = b[rowi].dup
					b[rowi].each_index {|diei|
						b[rowi][diei] = b[rowi][diei].dup
					}
				}
				b
			end
			def [](*x)
				if x.length == 2
					super(x[0])[x[1]]
				else
					super(*x)
				end
			end
		end
		def [](*x)
				@board[*x]
		end
		attr_reader :board
		def has_word?(word)
			if word =~ /q[^u]/i || word !~ /^[[:alpha:]]+$/
				return false
			end
			myword = []
			word.upcase.gsub(/QU/, 'Qu').scan(/[A-Z]u?/) {|x| myword.push x}
			if !myword.size.between?(3, @width * @width)
				return false
			end
			nb = @board.dup
			@board.find_letter(myword[0]).each {|coords|
				nb[*coords].concat '_'
				return true if nb.find_word(myword[1..-1], coords)
				nb[*coords].chop!
			}
			false
		end
		def to_s
			border = '+---' * @width + "+\n"
			rowfmt = "| %-2s" * @width + "|\n" + border
			string = border.dup
			@board.each {|row|
				string.concat format(rowfmt, *row)
			}
			string
		end
		def freeze
			freeze_board
			super
		end
		def dice_array
			dice = []
			@board.each {|row|
				row.each {|die|
					dice.push(die)
				}
			}
			dice
		end
		private
		def freeze_board
			@board.each {|row|
				row.each {|die|
					die.freeze
				}
				row.freeze
			}
			@board.freeze
		end
	end
	# Helper class to look up words with
	class Dict
		def initialize
			d = open('/usr/share/dict/words', 'r')
			@dict = Hash.new(false)
			def @dict.inspect
				"{Ooki}"
			end
			pat = /^[[:lower:]]+$/
			d.read.split("\n").each {|w|
				w.chomp!
				if w.size >= 3 && w =~ pat
					@dict[w] = true 
				end
			}
			d.close
			self
		end
		def [](word)
			@dict[word]
		end
		alias_method :has_word?, :[]
		def size
			@dict.size
		end
	end
end

if __FILE__ == $0
	def time
		t = Time.now
		[yield, Time.now - t]
	end
	print "Testing Bonkle::Dict.new... "
	$stdout.flush
	dict, t = time {Bonkle::Dict.new}
	printf("%d words in %f seconds\n", dict.size, t)
	puts "Testing Bonkle::Board..."
	print "Trying Bonkle::Board.new... "
	begin
		board = Bonkle::Board.new
	rescue
		puts $!
		exit 1
	else
		puts "success"
	end
	print "Trying Bonkle::Board#to_s... "
	begin
		print "\n" + board.to_s
	rescue
		puts $!
		exit 1
	end
	print <<-'EOF'
Trying Bonkle::Board#has_word?...
Enter a word to check for and get a "true" or "false" response.
Enter nothing or press EOF to end.
	EOF
	while (word = gets) != nil && word != "\n"
		word.chomp!
		puts "\t-> " + (board.has_word?(word) && dict[word]).to_s
	end
end
