#!/usr/local/bin/ruby
# Copyright (c) 2001 Brian Fundakowski Feldman
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: bonklenet.rb,v 1.3 2001/03/01 04:28:50 green Exp $
require 'bonkle'
require 'socket'
require 'thread'

module Bonkle
	module Message
		def Message.read(socket)
			m = socket.sysread(2)
			if m.size != 2
				return nil
			end
			socket.sysread(m.unpack('n')[0])
		end
		def Message.write(socket, message)
			socket.syswrite([message.size].pack('n').concat(message))
		end
		SUCCESS = 0
		FAILURE = 1
		@@ecodes = {
			0 => ' ',		# success
			1 => ' error: '	# failure
		}
		def Message.swrite(socket, ecode, message)
			message = ecode.to_s + @@ecodes[ecode] + message
			socket.syswrite([message.size].pack('n').concat(message))
		end
	end
	class Server
		class User
			[:name, :words, :status, :commandq].each {|att| attr_reader att}
			class MaxConn < Exception; end
			class NameUsed < Exception; end
			def initialize(sock, game)
				@socket = sock
				@game = game
				@name = ''
				@words = []
				@status = :init
				@starttime = Time.now
				@commandq = Queue.new
				@stop = false
				@thread = Thread.new {
					begin
						write("info Hello, please introduce yourself.")
						until @stop
							process_event
						end
					rescue EOFError => e
					rescue Exception => e
						$stderr.puts(e)
					ensure
						if @status == :run
							@game.remove_user(self)
						end
						Thread.exit
					end
				}
			end
			def points
				pn = 0
				@words.each {|word| pn += @game.points(word)}
				pn
			end
			private
			def select_wait(time)
				select([@socket], nil, nil, time)
			end
			def read
				Message.read(@socket)
			end
			def write(*args)
				Message.write(@socket, *args)
			end
			def swrite(*args)
				Message.swrite(@socket, *args)
			end
			def process_event
				until @commandq.empty?
					command = @commandq.shift
					case command[0]
					when :info
						write("info #{command[1]}")
					when :say
						write("say #{command[1]} #{command[2]}")
					when :shout
						write("shout #{command[1]} #{command[2]}")
					when :goodbye
						begin
							write("goodbye " + command[1])
						ensure
							@socket.close
							@stop = true
							if @status != :init
								@game.remove_user(self)
							end
							@status = :dead
							return
						end
					when :reset
						@words.empty
						swrite(Message::SUCCESS, "new game started")
					end
					return
				end

				if !select_wait(1)
					return
				end
				case @status
				when :init
					@message = read
					if @message =~ /^name\s*(.+)$/
						@name = $1
						if @name !~ /^[A-Za-z0-9_-]*$/ || @name.length > 20
							@name = ''
						end
					end
					if @name.empty?
						swrite(Message::FAILURE, "need valid name")
						return
					else
						begin
							@game.add_user(self)
						rescue MaxConn => e
							swrite(Message::FAILURE, e.to_s)
							@stop = true
							return
						rescue NameUsed => e
							swrite(Message::FAILURE, e.to_s)
							return
						end
					end
					@status = :run
					swrite(Message::SUCCESS, "name set")
				when :run
					@message = read
					case @message
					when /^timeleft(\s|$)/
						swrite(Message::SUCCESS,
						  format("%.4f", @game.endtime - Time.now))
					when /^fboard(\s|$)/
						swrite(Message::SUCCESS,
						  "pre-formatted Bonkle board:\n" +
						  @game.board.to_s)
					when /^boardsize(\s|$)/
						swrite(Message::SUCCESS, @game.board.width.to_s)
					when /^board(\s|$)/
						swrite(Message::SUCCESS,
						  @game.board.dice_array.join(' '))
					when /^lusers(\s|$)/
						swrite(Message::SUCCESS,
						  @game.usernames.join(' '))
					when /^lwords(\s+(.+)|$)/
						@game.crossout
						if $1.empty? || $2 == @name
							swrite(Message::SUCCESS, @words.sort.join(' '))
						else
							if @game.finished?
								msg = @game.user_do($2) {|u|
									if !u
										[Message::FAILURE, 'invalid user']
									else
										[Message::SUCCESS,
										  u.words.sort.join(' ')]
									end
								}
								swrite(*msg)
							else
								swrite(Message::FAILURE, 'game not finished')
							end
						end
					when /^points(\s+(.+)|$)/
						@game.crossout
						if $1.empty? || $2 == @name
							swrite(Message::SUCCESS, points.to_s)
						else
							if @game.finished?
								msg = @game.user_do($2) {|u|
									if !u
										[Message::FAILURE, 'invalid user']
									else
										[Message::SUCCESS, u.points.to_s]
									end
								}
								swrite(*msg)
							else
								swrite(Message::FAILURE, 'game not finished')
							end
						end
					when /^word\s+(.+)$/
						if @game.finished?
							swrite(Message::FAILURE, 'round already over')
							return
						end
						w = $1.chomp.downcase
						if !@game.board.has_word?(w)
							swrite(Message::FAILURE, 'word not found in board')
						elsif @words.include?(w)
							swrite(Message::FAILURE, 'word already recorded')
						elsif !@game.dict.has_word?(w)
							swrite(Message::FAILURE, 'word not recognized')
						else
							@words.push(w)
							swrite(Message::SUCCESS, 'word recorded')
						end
					when /^say\s+([^\s]+)\s(.+)$/
						msg = @game.user_do($1) {|u|
							if !u
								[Message::FAILURE, 'invalid user']
							else
								u.say(@name, $2)
								[Message::SUCCESS, 'message sent']
							end
						}
						swrite(*msg)
					when /^shout\s(.+)$/
						@game.each_user {|u|
							u.shout(@name, $1)
						}
						swrite(Message::SUCCESS, 'message shouted')
					else
						swrite(Message::FAILURE, 'invalid command')
					end
				end
			end
			def queue_event(event)
				@commandq.push(event)
			end
			public		# external interfaces
			def wait
				@thread.join
			end
			def goodbye(message)
				queue_event([:goodbye, message])
			end
			def say(from, message)
				queue_event([:say, from, message])
			end
			def shout(from, message)
				queue_event([:shout, from, message])
			end
		end
		attr_reader :game
		# uses a Hash for options
		def initialize(opts = {})
			@game = Game.new(:maxconn => opts[:maxconn])
			@stop = false
			opts[:host] ||= '0'
			opts[:port] ||= 31782
			@opts = opts
			@socket = TCPServer.new(opts[:host], opts[:port])
			@thread = Thread.new {
				begin
					until @stop
						process_events
					end
					@game.each_user {|user|
						user.goodbye('BonkleServer shut down')
					}
					@game.each_user {|user|
						user.wait
					}
					@socket.close
				rescue Exception => e
					$stderr.puts(e)
				ensure
					Thread.exit
				end
			}
		end
		def stop
			@stop = true
			@thread.join
		end
		def process_events
			if select([@socket], [], [], 1)
				begin
					s = @socket.accept
				rescue
					return
				end
				User.new(s, @game)
			end
		end
		private :process_events
	end
	class Game
		attr_reader :dict
		attr_reader :endtime
		attr_reader :mutex
		attr_accessor :gamelength
		def initialize(opts = {})
			@board = Bonkle::Board.new
			@dict = Bonkle::Dict.new
			@gamelength = 120
			@users = {}
			@mutex = Mutex.new
			@starttime = Time.now
			@endtime = @starttime + @gamelength
			opts[:maxconn] ||= 5
			if !opts[:points]
				pn = opts[:points] = Hash.new(11)
				pn[3] = 1
				pn[4] = 1
				pn[5] = 2
				pn[6] = 3
				pn[7] = 5
				pn[8] = 7
			end
			@opts = opts
		end
		def reset
			@mutex.synchronize {
				@board.reset
				@starttime = Time.now
				@endtime = @starttime + @gamelength
				@crossedout = false
			}
		end
		def crossout
			@mutex.synchronize {
				if finished? && !@crossedout
					@crossedout = true
					users = @users.collect {|name, user| [user, user.words.dup]}
					users.each {|user|
						users.each {|user2|
							next if user[0].eql?(user2[0])
							user2[1].each {|word2|
								user[0].words.delete(word2)
							}
						}
					}
				end
			}
		end
		def points(word)
			@opts[:points][word.size]
		end
		def user_do(name)
			@mutex.synchronize {
				yield @users[name]
			}
		end
		def add_user(user)
			@mutex.synchronize {
				if @users.size == @opts[:maxconn]
					raise(user.type::MaxConn,
					  "maximum number of connections exceeded")
				elsif @users.has_key?(user.name)
					raise(user.type::NameUsed, "user name in use")
				else
					@users[user.name] = user
				end
			}
		end
		def remove_user(user)
			@mutex.synchronize {
				@users.delete(user.name)
			}
		end
		def each_user
			@mutex.synchronize {@users.dup}.each {|name, user| yield user}
		end
		# Through wonderful use of GC-enabled language, this isn't necessary ;)
		def each_user_locked
			@mutex.synchronize {@users.each {|name, user| yield user}}
		end
		def usernames
			@mutex.synchronize {@users.collect {|name, user| name}}
		end
		def has_word?(word)
			@board.has_word?(word)
		end
		def finished?
			Time.now >= @endtime
		end
		def board
			@board
		end
	end
	class Client
		def initialize(opts = {})
			opts[:host] ||= 'localhost'
			opts[:port] ||= 31782
			@socket = TCPSocket.new(opts[:host], opts[:port])
		end
		def write(msg)
			Message::write(@socket, msg)
		end
		def read
			Message::read(@socket)
		end
		def tryread
			Message::read(@socket) if select([@socket], nil, nil, 0)
		end
	end
end
